"""Small utilities that play around with propositions.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)

"""
from __future__ import nested_scopes
import operator
import structure
import copy



def getIdentifiers(tree):
    """Return all the unique identifiers of a propositional parse
    tree."""
    children_identifiers = mappend(getIdentifiers, tree.children)
    if isIdentifierNode(tree):
        return uniq([tree.attr] + children_identifiers)
    return uniq(children_identifiers)


def implicationToDisjunction(impl_node):
    """Given an implication, returns an equivalent disjunction.
    Otherwise, just returns impl_node.

    Not destructive."""
    if not isImplicationNode(impl_node): return impl_node
    return structure.OrNode(structure.NotNode(impl_node[0]),
                            impl_node[1])


def makeFullStateTables(identifiers):
    """Given a list of identifier names, returns a list of all possible
    environments for these identifiers.

    Be careful: this is very expensive and is guaranteed to return
    2**(len(identifiers)) elements!  If I have time, I should probably
    make this an iterative algorithm, or avoid constructing this
    altogether."""
    if len(identifiers) == 0: return [{}]
    partial_state_tables_true = makeFullStateTables(identifiers[1:])
    partial_state_tables_false = copy.deepcopy(partial_state_tables_true)
    for i in range(len(partial_state_tables_true)):
        partial_state_tables_true[i][identifiers[0]] = structure.TRUE_VALUE
        partial_state_tables_false[i][identifiers[0]] = structure.FALSE_VALUE
    return partial_state_tables_true + partial_state_tables_false


def propositionToStates(proposition_tree):
    """Given a proposition_tree, return the list of states such that
    evaluation in any of of them causes the proposition to evaluate
    as a true statement."""
    ids = getIdentifiers(proposition_tree)
    all_states = makeFullStateTables(ids)
    true_states = filter(lambda s: (proposition_tree.eval(s) ==
                                    structure.TRUE_VALUE),
                         all_states)
    return true_states
                         

def propositionToDisjunction(prop_tree):
    pass  ## fixme

def propositionToConjunction(prop_tree):
    pass  ## fixme


## def deMorganize(node):
##     pass  ## fixme!


######################################################################
def isIdentifierNode(node):
    return isinstance(node, structure.IdentifierNode)


def isAndNode(node):
    return isinstance(node, structure.AndNode)


def isOrNode(node):
    return isinstance(node, structure.OrNode)


def isNotNode(node):
    return isinstance(node, structure.NotNode)


def isImplicationNode(node):
    return isinstance(node, structure.ImplicationNode)


def isEqualityNode(node):
    return isinstance(node, structure.EqualityNode)
######################################################################


def mappend(function, l):
    """Perform a mapping append."""
    if not l: return []
    return reduce(operator.add, map(function, l))


def uniq(L):
    """Return all the unique elements of l as a list"""
    dict = {}
    for element in L:
        dict[element] = 1
    return dict.keys()


def deepTransform(function, tree):
    new_tree = copy.deepcopy(tree)
    new_children = map(lambda t: deepTransform(function, t), tree.children)
    new_tree.children = new_children
    return function(new_tree)
