"""These classes provide support to represent and evaluate
propositional logic.

Danny Yoo (dyoo@hkn.eecs.berkeley.edu)
"""
from __future__ import nested_scopes


TRUE_VALUE = 'T'
FALSE_VALUE = 'F'


def nodeEval(node, env=None):
    """A functional wrapper that simply calls a node's eval()
    function."""
    return node.eval(env)


class Node:
    def __init__(self, type, attr, left=None, right=None, children=()):
        self.type = type
        self.attr = attr
        self.children = list(children)
        if left: self.children.append(left)
        if right: self.children.append(right)


    def __getitem__(self, i):
        return self.children[i]

 
    def isLeaf(self):
        return len(self.children) == 0
 

    def __str__(self):
        if self.isLeaf(): return "%s" % self.attr
        child_list = map(str, self.children)
        if self.attr == None:
            return "%s[%s]" % (self.type, ', '.join(child_list))
        else:
            return "%s[%s]" % (self.attr, ', '.join(child_list))


    def eval(self, env):
        """Evaluate a node, given a particular env binding that binds
        identifiers to values, and return that value."""
        print "I haven't been defined yet.  Fixme!"



class ImplicationNode(Node):
    def __init__(self, left, right):
        Node.__init__(self, type='implication', attr=None,
                      left=left, right=right)

    def eval(self, env):
        if self[0].eval(env) == FALSE_VALUE:
            return TRUE_VALUE
        if self[1].eval(env) == TRUE_VALUE:
            return TRUE_VALUE
        return FALSE_VALUE


class EqualityNode(Node):
    def __init__(self, left, right):
        Node.__init__(self, type='equality', attr=None,
                      left=left, right=right)

    def eval(self, env):
        return primitiveEquals(self[0].eval(env), self[1].eval(env))

    
class AndNode(Node):
    def __init__(self, *nodes):
        Node.__init__(self, type='and', attr=None,
                      children=nodes)

    def eval(self, env):
        values = map(lambda c: nodeEval(c, env), self.children)
        return primitiveAnd(*values)
    

class OrNode(Node):
    def __init__(self, *nodes):
        Node.__init__(self, type='or', attr=None,
                      children=nodes)

    def eval(self, env):
        values = map(lambda c: nodeEval(c, env), self.children)
        return primitiveOr(*values)


class NotNode(Node):
    def __init__(self, node):
        Node.__init__(self, type='not', attr=None,
                      children=[node])

    def eval(self, env):
        value = nodeEval(self[0], env)
        return primitiveNot(value)


class IdentifierNode(Node):
    def __init__(self, identifier):
        Node.__init__(self, type='identifier', attr=identifier)

    def eval(self, env):
        return env[self.attr]
        ## fixme!  I want to implement nestable environments, and make
        ## this more robust.


class TrueNode(Node):
    def __init__(self):
        Node.__init__(self, type='constant', attr=TRUE_VALUE)

    def eval(self, env):
        return TRUE_VALUE
    

class FalseNode(Node):
    def __init__(self):
        Node.__init__(self, type='constant', attr=FALSE_VALUE)

    def eval(self, env):
        return FALSE_VALUE
    

def primitiveAnd(*values):
    for v in values:
        if v == FALSE_VALUE:
            return FALSE_VALUE
    return TRUE_VALUE


def primitiveOr(*values):
    for v in values:
        if v == TRUE_VALUE:
            return TRUE_VALUE
    return FALSE_VALUE


def primitiveNot(value):
    if value == TRUE_VALUE:
        return FALSE_VALUE
    return TRUE_VALUE


def primitiveEquals(lhs, rhs):
    if lhs == rhs: return TRUE_VALUE
    return FALSE_VALUE
